#include <sys/stat.h>
#include <sys/types.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <limits.h>
#include <math.h>

#include "common.h"
#include "macros.h"
#include "failures.h"
#include "conf.h"

static const char *parser_errors[] = {
	"section not closed properly",
	"key name not found",
	"wrong key name format",
	"key out of the section",
	"empty key value",
	"equality sign is not found",
	"can't extract integer value",
	"can't extract float value",
	"can't extract string value",
	"can't extract boolean value",
	"unexpected token"
	"undefined section"
};

static void skip_blanks(char **cur)
{
	char *pos;

	pos = *cur;
	while (isblank(*pos))
		++pos;

	*cur = pos;
}

static void skip_key_name(char **cur)
{
	char *pos;

	pos = *cur;

	if (isalpha(*pos) || *pos == '_') {
		do {
			++pos;
		} while (isalnum(*pos) || *pos == '_');
	}

	*cur = pos;
}

static void remove_trailing_spaces(char *ptr, size_t len)
{
	while (len && isspace(ptr[len-1]))
		ptr[--len] = 0;
}

static int yesno(const char *str)
{
	unsigned int i;
	struct {
		const char	*str;
		int		res;
	} bool_values[] = {
		{ "yes",	1 },
		{ "true",	1 },
		{ "1",		1 },
		{ "no",		0 },
		{ "false",	0 },
		{ "0",		0 },
		{ NULL,		-1 }
	};

	for (i = 0; bool_values[i].str != NULL; i++) {
		if (!strcasecmp(bool_values[i].str, str))
			return bool_values[i].res;
	}

	return -1;
}

static void error_handler(struct config_file *cf, config_error_type_t type)
{
	const char *msg;

	msg = parser_errors[type];

	if (cf->error_handler != NULL)
		cf->error_handler(cf, type, msg);
	else
		fprintf(stderr, "%s %s:%d\n", msg, cf->file_name, cf->line_number);
}

static int extract_integer(struct config_file *cf, struct cmd *cmd, char **cur)
{
	long int num;
	char *beg;

	beg = *cur;
	errno = 0;

	num = strtol(beg, cur, 0);

	if ((errno == ERANGE && (num == LONG_MAX || num == LONG_MIN)) ||
	     errno == EINVAL || errno != 0) {
		error_handler(cf, CONF_ERR_EXTR_INT);
		return -1;
	}

	cmd->value.integer = num;

	return 0;
}

static int extract_real(struct config_file *cf, struct cmd *cmd, char **cur)
{
	float fl;
	char *beg;

	beg = *cur;
	fl = strtof(beg, cur);

	if (errno == ERANGE) {
		error_handler(cf, CONF_ERR_EXTR_REAL);
		return -1;
	}

	if (isnan(fl) || isinf(fl)) {
		error_handler(cf, CONF_ERR_EXTR_REAL);
		return -1;
	}

	cmd->value.real = fl;
	return 0;
}

static int extract_string(struct config_file *cf, struct cmd *cmd, char **cur)
{
	char *beg, *ptr;

	ptr = *cur;

	if (*ptr != '\"') {
		beg = ptr;
		while (*ptr != '\0' && !isblank(*ptr))
			++ptr;
		*cur = ptr;
	} else {
		beg = ++ptr;
		while (*ptr != '\0' && *ptr != '\"') {
			if (ptr[0] == '\\' && ptr[1] != '\0')
				ptr += 2;
			else
				++ptr;
		}

		if (*ptr != '\"') {
			error_handler(cf, CONF_ERR_EXTR_STR);
			return -1;
		}
		*cur = ptr+1;
	}

	*ptr = 0;
	cmd->value.str = beg;

	return 0;
}

static int extract_bool(struct config_file *cf, struct cmd *cmd, char **cur)
{
	char *beg, *end, ch;

	beg = end = *cur;

	while (isdigit(*end) || isalpha(*end))
		++end;

	ch = *end;
	*end = 0;

	switch (yesno(beg)) {
	case 1:
		cmd->value.bool = TRUE;
		break;
	case 0:
		cmd->value.bool = FALSE;
		break;
	default:
		error_handler(cf, CONF_ERR_EXTR_BOOL);
		return -1;
	}

	*end = ch;
	*cur = end;

	return 0;
}

static int handle_key(struct config_file *cf, char *str)
{
	struct cmd cmd;
	char *key, *cur;
	unsigned int i;
	int ret;

	/* A key must be in one section. */
	if (cf->current == NULL) {
		error_handler(cf, CONF_ERR_KEY_OUT_SECT);
		return -1;
	}

	key = cur = str;
	skip_key_name(&cur);
	/* A wrong key format. */
	if (key == cur) {
		error_handler(cf, CONF_ERR_KEY_NAME_FORMAT);
		return -1;
	}

	skip_blanks(&cur);
	if (*cur != '=') {
		error_handler(cf, CONF_ERR_EQU_NOT_FOUND);
		return -1;
	}

	*cur++ = 0;
	remove_trailing_spaces(key, strlen(key));
	/* Is key value empty? */
	if ((strlen(key)) == 0) {
		error_handler(cf, CONF_ERR_KEY_NAME);
		return -1;
	}
	/* Come to the value. */
	skip_blanks(&cur);
	/* Matching the name. */
	for (i = 0; cf->current->keys[i].name != NULL; i++) {
		if (!STREQ(key, cf->current->keys[i].name))
			continue;
		break;
	}

	/* Don't find a right option. Just skip the unknown key. */
	if (cf->current->keys[i].name == NULL)
		return 0;

	cmd.section = cf->current->name;
	cmd.key = key;
	/* Extract a key value according to the registered type. */
	switch (cf->current->keys[i].val_type) {
	case CONF_TYPE_INTEGER:
		ret = extract_integer(cf, &cmd, &cur);
		break;
	case CONF_TYPE_REAL:
		ret = extract_real(cf, &cmd, &cur);
		break;
	case CONF_TYPE_STRING:
		ret = extract_string(cf, &cmd, &cur);
		break;
	case CONF_TYPE_BOOL:
		ret = extract_bool(cf, &cmd, &cur);
		break;
	default:
		break;
	}

	if (ret != 0)
		return ret;

	skip_blanks(&cur);
	if (*cur != 0) {
		error_handler(cf, CONF_ERR_UNEXPECT_TOKEN);
		return -1;
	}
	/* Call handler. */
	ret = 0;
	if (cf->current->keys[i].handler != NULL)
		ret = cf->current->keys[i].handler(&cmd, cf->opaque);

	return ret;
}

static int handle_section(struct config_file *cf, const char *str)
{
	char *cur;
	unsigned int i;

	cur = strchr(str, ']');
	/* Unclosed section definition. Call error handler if it is available
	 * or terminate. */
	if (cur == NULL) {
		error_handler(cf, CONF_ERR_SECT_NAME);
		return -1;
	}

	*cur++ = 0;
	cf->current = NULL;
	/* Find a section by name. */
	for (i = 0; cf->sections[i].name != NULL; i++) {
		if (!STREQ(cf->sections[i].name, str))
			continue;

		cf->current = &cf->sections[i];
		break;
	}
	/* Unknown section. */
	if (cf->current == NULL) {
		error_handler(cf, CONF_ERR_UNDEF_SECTION);
		return -1;
	}

	skip_blanks(&cur);
	if (*cur != 0) {
		error_handler(cf, CONF_ERR_UNEXPECT_TOKEN);
		return -1;
	}

	if (cf->current->handler != NULL)
		cf->current->handler(cf->opaque);

	return 0;
}

static int handle_line(struct config_file *cf, char *str)
{
	int ret;

	switch (*str) {
	case '[':
		++str;
		ret = handle_section(cf, str);
		break;
	default:
		ret = handle_key(cf, str);
		break;
	}
	
	return ret;
}

int config_parse(struct config_file *cf)
{
	char buf[PATH_MAX];
	char *line;
	size_t n;
	int ret;

	return_val_if_fail(cf != NULL, -1);

	ret = 0;
	while ((fgets(buf, sizeof(buf), cf->fp)) != NULL) {

		n = strlen(buf);

		if (buf[n-1] == '\n') 
			buf[n-1] = 0;
		else
			continue;

		++cf->line_number;

		line = buf;
		skip_blanks(&line);

		if (*line == 0)
			continue;

		if (*line == '#' || *line == ';')
			continue;

		ret = handle_line(cf, line);
		if (ret != 0)
			break;
	}

	return ret;
}

struct config_file *config_create(const char *file_name, struct section_option *sections,
				  void *opaque)
{
	struct stat st;
	struct config_file *cf;
	FILE *fp;

	return_val_if_fail(file_name != NULL, NULL);
	return_val_if_fail(sections != NULL, NULL);

	if ((stat(file_name, &st) != 0) || !S_ISREG(st.st_mode))
		return NULL;

	fp = fopen(file_name, "r");
        if (fp == NULL)
        	return NULL;

	cf = malloc(sizeof(*cf));
	if (cf == NULL)
		goto nomem;

	memset(cf, 0, sizeof(*cf));

	cf->file_name = strdup(file_name);
	if (cf->file_name == NULL)
		goto nomem2;

	cf->fp = fp;
	cf->sections = sections;
	cf->opaque = opaque;

	return cf;

nomem:
	free(cf);
nomem2:
	fclose(fp);

	return NULL;
}

void config_destroy(struct config_file **cf)
{
	return_if_fail(cf != NULL);
	return_if_fail(*cf != NULL);

	if ((*cf)->file_name != NULL)
		free((*cf)->file_name);
	fclose((*cf)->fp);

	free((*cf));
	*cf = NULL;
}

